#include <Firmata.h>

#define DATAMODETEXT 1

#ifdef DATAMODETEXT
int datamode=0;
#endif

/* IO Pin Mapping.
 ** Digital Pin 0 - RX
 ** Digital Pin 1 - TX
 ** Digital Pin 2-7 - SNES APU D2-7
 ** Digital Pin 8 - SNES APU address 0
 ** Digital Pin 9 - SNES APU address 1
 ** Digital Pin 10 - SNES APU /RD
 ** Digital Pin 11 - SNES APU /WR
 ** Digital Pin 12-13 - SNES APU D0-1
 ** Analog Pin 0 (AKA Digital Pin 14) - SNES APU /RESET
 ** Vcc - SNES APU address 6, SNES APU Vcc, SNES APU audio Vcc
 ** Gnd - SNES APU Gnd, SNES APU audio Gnd.
 */

unsigned char port0state;

void setup()
{
  PORTB = 0x0F;
  DDRB = 0x0F;
  PORTC = 0x01;
  DDRC = 0x01;
  Serial.begin(115200);
  Serial.println("SPC700 DATA LOADER V1.0");
}


unsigned char readdata(unsigned char address)
{
  unsigned char data;

  DDRB = 0x0F;
  DDRD &= ~0xFC;
  PORTB &= ~0x03;
  PORTB |= (address & 0x03);

  PORTB &= ~0x04;
  __asm__ __volatile__ ("nop");
  __asm__ __volatile__ ("nop");
  data=((PINB&0x30)>>4) | (PIND&0xFC);
  PORTB |= 0x04;
  return data;
}

void writedata(unsigned char address, unsigned char data)
{
  DDRB = 0x3F;
  DDRD |= 0xFC;

  PORTD &= ~0xFC;
  PORTD |= (data & 0xFC);

  PORTB &= ~0x33;
  PORTB |= (address & 0x03);
  PORTB |= ((data & 0x03)<<4);
  PORTB &= ~0x08;
  __asm__ __volatile__ ("nop");
  __asm__ __volatile__ ("nop");
  PORTB |= 0x08;

}

/*
unsigned char program_counter=0xFF;
 uint32_t delay_count=0;
 unsigned char Script700[64];
 unsigned char Script700WorkArea[8];
 unsigned char Script700RAM[64];*/

void ProcessScript700()
{
  // --- TODO:  Implement script 700 processing, and script700 uploading procedures.
  // --- Limitations.  Only ~512 bytes of working ram/code space to work with.
  // --- No direct access to the APU ram/registers in any way, shape or form,
  // --- other than IO ports F4-F7.
  /*
  if(program_counter==0xFF)
   return;
   
   if(delay_count)
   {
   if(delay_count>=32)
   {
   delay_count-=32;
   delayMicroseconds(16);
   }
   else
   {
   delayMicroseconds(delay_count>>1);
   delay_count=0;
   }
   return;
   }
   
   switch(Script700[program_counter&0x3F]&0x0F)
   {
   case 0:  //Wait
   delay_count = Script700[program_counter+1];
   delay_count |= Script700[program_counter+2]<<8;
   delay_count |= Script700[program_counter+1] << 16;
   delay_count |= Script700[program_counter+1] << 24;
   program_counter+=5;
   break;
   case 1:  //Move
   break;
   default:
   program_counter=0xFF;
   break;
   }*/

}

unsigned char readserial()
{
  while (Serial.available() == 0)
    ProcessScript700();
  return Serial.read();
}


#ifdef DATAMODETEXT
unsigned char readhex()
{
  unsigned char data[2];
  data[0]=readserial();
  data[1]=readserial();
  if((data[0]>='0')&&(data[0]<='9'))
    data[0]-='0';
  else if ((data[0]>='A')&&(data[0]<='F'))
  {
    data[0]-='A';
    data[0]+=10;
  }
  else if ((data[0]>='a')&&(data[0]<='f'))
  {
    data[0]-='a';
    data[0]+=10;
  }
  else
    data[0]=0;

  if((data[1]>='0')&&(data[1]<='9'))
    data[1]-='0';
  else if ((data[1]>='A')&&(data[1]<='F'))
  {
    data[1]-='A';
    data[1]+=10;
  }
  else if ((data[1]>='a')&&(data[1]<='f'))
  {
    data[1]-='a';
    data[1]+=10;
  }
  else
    data[1]=0;

  return (data[0]<<4)|data[1];
}
#endif

void startapuwrite(uint16_t address, unsigned char *data, int len)
{
  int i;
  digitalWrite(14,LOW);
  delay(1);
  digitalWrite(14,HIGH);

  while(readdata(0)!=0xAA);
  while(readdata(1)!=0xBB);

  writedata(3,address>>8);
  writedata(2,address&0xFF);
  writedata(1,1);
  writedata(0,0xCC);
  while(readdata(0)!=0xCC);
  for(i=0;i<len;i++)
  {
    writedata(1,data[i]);
    writedata(0,i&0xFF);
    while(readdata(0)!=(i&0xFF));
  }
}

/* // --- Implemented for faster emulator linking, on at least the initial code upload.
 // The code will be skipped by usual fashion, but data recorded for to do the initial upload.
 // Once the data is uploaded, all 8/16 bit read/writes will be passed on to the emulator to control.
 // startapuwrite is not commented out, because it is being used for spc file uploading process.
 void continueapuwrite(uint16_t address, unsigned char *data, int len)
 {
 int i;
 i=readdata(0);
 i+=2;
 writedata(3,address>>8);
 writedata(2,address&0xFF);
 writedata(1,1);
 writedata(0,i&0xFF);
 while(readdata(0)!=(i&0xFF));
 for(i=0;i<len;i++)
 {
 writedata(1,data[i]);
 writedata(0,i&0xFF);
 while(readdata(0)!=(i&0xFF));
 }
 }
 
 void executeapu(uint16_t address)
 {
 int i;
 i=readdata(0);
 i+=2;
 writedata(3,address>>8);
 writedata(2,address&0xFF);
 writedata(1,1);
 writedata(0,i&0xFF);
 //while(readdata(0)!=(i&0xFF)); //Commented just in case executed code clears the i+2 result immediately.
 }*/

void APU_Wait(unsigned char address, unsigned char data)
{
  while(readdata(address)!=data);
}

unsigned char DSPdata[16] =
{  //For loading the 128 byte DSP ram. DO NOT CHANGE.
  0xC4, 0xF2, 0x64, 0xF4, 0xD0, 0xFC, 0xFA, 0xF5, 0xF3, 0xC4, 0xF4, 0xBC, 0x10, 0xF2, 0x2F, 0xB7, 
} 
;


void loop()
{
  uint16_t i;
  long time;
  int checksum;
  unsigned char address, data;
  switch(readserial())
  {
  case 'D':    //Set datamode between ascii and binary.  In ascii mode,  each data byte is transmitted
    //as ascii hexadecimal notation.  In binary mode, the byte transmitted on the serial
    //line is taken literally for what it is.
#ifdef DATAMODETEXT
    datamode=readserial()-'0';
    if(datamode)
      Serial.println("Data mode set to Text");
#else
    readserial();  //Ignore this byte, but keep it here anyways.
#endif
    break;


  case 'Q':  //Read all 4 IO ports in rapid fire succession.
#ifdef DATAMODETEXT
    if(!datamode)
    {
#endif
      data=readdata(0);
      
      Serial.write(data);
      data=readdata(1);
      Serial.write(data);
      data=readdata(2);
      Serial.write(data);
      data=readdata(3);
      Serial.write(data);
#ifdef DATAMODETEXT
    }
    else
    {
      Serial.print("Data on All 4 ports = ");
      data=readdata(0);
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      data=readdata(1);
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      data=readdata(2);
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      data=readdata(3);
      if(data<16)
        Serial.print(0,HEX);
      Serial.println(data,HEX);
    }
#endif
    break;
  case 'R':  //Read a single IO port.  Takes an address as a parameter, returns one byte.
    address=readserial()-'0';
    data=readdata(address);
#ifdef DATAMODETEXT
    if(datamode)
    {
      Serial.print("Data at Address ");
      Serial.print(address,DEC);
      Serial.print(" is ");
      if(data<16)
        Serial.print(0,HEX);
      Serial.println(data,HEX);
    }
    else
#endif
      Serial.write(data);
    break;
  case 'r':  //Read and returns 2 consecutive IO ports.
    address=readserial()-'0';
    data=readdata(address+1);
#ifdef DATAMODETEXT
    if(datamode)
    {
      Serial.print("Data at Address ");
      Serial.print(address,DEC);
      Serial.print(" is ");
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      data=readdata(address);
      if(data<16)
        Serial.print(0,HEX);
      Serial.println(data,HEX);
    }
    else
    {
#endif
      Serial.write(data);
      data=readdata(address);
      Serial.write(data);
#ifdef DATAMODETEXT
    }
#endif
    break; 
  case 'S':
    Serial.println("SPC700 DATA LOADER V1.0");
    break;
  case 's':  //Reset the Audio Processing Unit
#ifdef DATAMODETEXT
    if(datamode)
      Serial.println("APU Reset");
#endif
    //Try the Analog line first.
    digitalWrite(14,LOW);
    delay(1);
    digitalWrite(14,HIGH);
    if(readdata(0)!=0xAA)  //Check for reset.
    {
      //And now try the OR gate, just in case the user has that hooked up instead.
      digitalWrite(10,LOW);
      digitalWrite(11,LOW);
      delay(1);
      digitalWrite(11,HIGH);
      digitalWrite(10,HIGH);
    }
    break;
  case 'f':  //Set the expected state of PORT 0 for the next write.
    port0state=readserial();
    break;
  case 'F':  //Upload SPC data at serial port speed.
    uint16_t ram_addr;
    uint16_t data_size;

    ram_addr=readserial()<<8;
    ram_addr|=readserial();
    data_size=readserial()<<8;
    data_size|=readserial();
    writedata(1,1);
    writedata(2,ram_addr&0xFF);
    writedata(3,ram_addr>>8);
    i=readdata(0);
    i+=2;
    writedata(0,i&0xFF);
    while(readdata(0)!=(i&0xFF));
    port0state=0;
    for(i=0;i<data_size;i++)
    {
      data=readserial();
      writedata(1,data);
      writedata(0,port0state);
      //while(readdata(0)!=port0state);
      port0state++;
    }
    Serial.write(1);

    break;

  case 'G':  //Upload the DSP register, and the first page of SPC data at serial port speed.
    startapuwrite(0x0002,DSPdata,16);
    writedata(2,0x02);
    writedata(3,0x00);
    writedata(1,0x00);
    writedata(0,0x11);
    while(readdata(0)!=0x11);
    port0state=0;
    for(i=0;i<128;i++)
    {
      data=readserial();
      writedata(1,data);
      writedata(0,port0state);
      if(i<127)
        while(readdata(0)!=port0state);
      port0state++;
    }
    while(readdata(0)!=0xAA);
    port0state=0;
    writedata(2,0x02);
    writedata(3,0x00);
    writedata(1,0x01);
    writedata(0,0xCC);
    while(readdata(0)!=0xCC);
    for(i=0;i<256;i++)
    {
      data=readserial();
      if(i<2)
        continue;
      if(i>=0xF0)
        continue;
      writedata(1,data);
      writedata(0,port0state);
      while(readdata(0)!=port0state);
      port0state++;
    }
    Serial.write('F');
    break;
  case 'W':  //Write a single IO port.
    address=readserial()-'0';
#ifdef DATAMODETEXT
    if(datamode)
      data=readhex();
    else
#endif
      data=readserial();
    switch(address>>2)
    {
      case 1:
        while(readdata(1)!=((readdata(0)+1)&0xFF));  //Handle a blazeon track change.
        /*
        ** Blazeon uses the following port communication system.
        ** When Port 1 is Equal to Port 0 + 1, then it is safe to write a new value to any of the 4 ports.
        ** When Port 0 is Equal to Port 1 + 1, then the SPC side driver is reading the ports, then clearing
        ** the ports.  As a result, if we write during the latter time, it ends up being a race condition to
        ** get the write in the specific port, before that port is read out.
        ** Immediately following the readout of the ports, the in ports are automatically cleared and will read
        ** as 0 on SPC side.
        **
        ** There is absolutely no SPC reloading code stored in a Blazeon SPC.  Once it is running, it runs forever,
        ** with its preprogrammed songs/sfxs. Writing to any port changes songs/sfxs played instantly during the
        ** former condition.
        */
      case 0:
      default:
        /*
        ** Default, is just write the ports as is.
        */
        writedata(address,data);
    }
#ifdef DATAMODETEXT
    if(datamode)
    {
      Serial.print("Writing 8bit ");
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      Serial.print(" to address ");
      Serial.println(address,DEC);
    }
#endif
    break;
  case 'w':  //Write 2 IO consecutive IO ports.
    address=readserial()-'0';
#ifdef DATAMODETEXT
    if(!datamode)
    {
#endif
      data=readserial();
      writedata(address+1,data);
      data=readserial();
      writedata(address,data);
#ifdef DATAMODETEXT
    }
    else
    {
      data=readhex();
      writedata(address+1,data);
      Serial.print("Writing 16bit ");
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      data=readhex();
      writedata(address,data);
      Serial.print(" ");
      if(data<16)
        Serial.print(0,HEX);
      Serial.print(data,HEX);
      Serial.print(" to address ");
      Serial.println(address,DEC);
    }
#endif
    break;

  }
}
